/* Copyright (c) 2004, Jasper van de Gronde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above
 *     copyright notice, this list of conditions and the
 *     following disclaimer.
 *  2. Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. Neither the name of the author nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ringbuffer.h"
#include "interlocked.h"
#include <memory.h>

lockfree::untyped_ringbuffer::untyped_ringbuffer(size_t size_p) :
  store(new char[size_p]), storesize(size_p)
{
  clear();
}

lockfree::untyped_ringbuffer::~untyped_ringbuffer() {
  delete store;
}

size_t lockfree::untyped_ringbuffer::write(const char* data, size_t size_p) {
  size_t max_size = distance(write_pos, read_pos);
  if ( max_size == 0 ) {
    max_size = storesize;
  }
  --max_size; // write pointer isn't allowed to make itself >= read_pos
  // The alternative is that the buffer somehow remembers wether the write pos is in
  // front of the read pos or not, but this seemed easier and usually you won't be
  // filling the buffer to the last byte anyway (or you won't care about one byte).
  const size_t size = size_p > max_size ? max_size : size_p;

  const size_t new_write_pos = (write_pos+size) % storesize;

  if ( new_write_pos >= write_pos ) {
    memcpy(store + write_pos, data, size);
  }
  else {
    const size_t first_size = storesize - write_pos;
    memcpy(store + write_pos, data, first_size);
    memcpy(store, data + first_size, size - first_size);
  }

  interlocked_exchange(&write_pos, new_write_pos);

  return size_p;
}

size_t lockfree::untyped_ringbuffer::read(char* data, size_t max_size_p) {
  const size_t max_size = distance(read_pos, write_pos);
  const size_t size = max_size > max_size_p ? max_size_p : max_size;

  const size_t new_read_pos = (read_pos+size) % storesize;

  if ( new_read_pos >= read_pos ) {
    memcpy(data, store + read_pos, size);
  }
  else {
    const size_t first_size = storesize - read_pos;
    memcpy(data, store + read_pos, first_size);
    memcpy(data + first_size, store, size - first_size);
  }

  interlocked_exchange(&read_pos, new_read_pos);

  return size;
}

void lockfree::untyped_ringbuffer::clear() {
  read_pos = 0;
  write_pos = 0;
}

size_t lockfree::untyped_ringbuffer::distance(size_t ref, size_t pos) {
  return pos >= ref ? pos-ref : storesize-(ref-pos);
  //long dist = pos-ref;
  //return dist >= 0 ? dist : storesize+dist;
}
/*
long lockfree::untyped_ringbuffer::closest(long ref, long a, long b) {
  long da = distance(ref, a);
  long db = distance(ref, b);
  return da <= db ? a : b;
}
*/